import re
from sqlalchemy import inspect
from sqlalchemy.orm.state import AttributeState
from sqlalchemy.util import classproperty

name_re = re.compile('([a-z]|^)([A-Z])')


def name_subber(match):
    if match.group(1):
        start = match.group(1)+'_'
    else:
        start = ''
    return start+match.group(2).lower()


def de_hump(cls):
    "SomeThing -> some_thing"
    return name_re.sub(name_subber,cls.__name__)


def comparable_attributes(obj):
    state = inspect(obj)
    relationships = state.mapper.relationships
    for name, attr in state.attrs.items():
        if name in relationships:
            continue
        yield name, attr.value


class Common(object):
    
    @classproperty
    def __tablename__(cls):
        return de_hump(cls)
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return False
        other_attrs = inspect(other).attrs
        for name, value in comparable_attributes(self):
            if value != other_attrs[name].value:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        content = []
        for name, value in sorted(comparable_attributes(self)):
            if value is not None:
                content.append('%s=%r' % (name, value))
        return '%s(%s)' % (self.__class__.__name__, ', '.join(content))

    __str__ = __repr__


def compare_common(x, y, context):

    ignore_fields = context.get_option('ignore_fields', ())

    if type(x) is not type(y):
        return compare_simple(x, y, context)

    args = []

    for obj in x, y:
        attrs = {}
        for name, value in comparable_attributes(obj):
            if name in ignore_fields:
                continue
            attrs[name] = value
        args.append(attrs)

    if ignore_fields and args[0]==args[1]:
        return

    args.append(context)
    args.append(x)

    return _compare_mapping(*args)


try:
    from testfixtures.comparison import (
        register, _compare_mapping, compare_simple
    )
except ImportError: # pragma: no cover
    pass
else:
    register(Common, compare_common)



from unittest import TestCase
from sqlalchemy.orm import relationship, joinedload
from mortar_mixins.common import Common
from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from testfixtures import compare
from mortar_rdb import get_session
from mortar_rdb.testing import register_session


class SetupModels(object):

    def setUp(self):
        register_session(transactional=False)

        self.Base = declarative_base()

        class Model(Common, self.Base):
            id =  Column(Integer, primary_key=True)
            value = Column(Integer)

        class AnotherModel(Common, self.Base):
            id = Column(Integer, primary_key=True)
            attr = Column(Integer)
            other_id = Column(Integer, ForeignKey('model.id'))
            other = relationship("Model", backref='another')

        self.Model = Model
        self.AnotherModel = AnotherModel

        self.session = get_session()
        self.addCleanup(self.session.rollback)
        self.Base.metadata.create_all(self.session.bind)


class CommonTests(SetupModels, TestCase):

    def test_table_name(self):
        compare(self.Model.__table__.name, 'model')

    def test_eq_wrong_type(self):
        self.assertFalse(self.Model() == object())

    def test_eq_wrong_model_type(self):
        class OtherModel(Common, self.Base):
            id = Column(Integer, primary_key=True)
        self.assertFalse(self.Model(id=1) == OtherModel(id=1))

    def test_eq_different(self):
        self.assertFalse(self.Model(id=1) == self.Model(id=2))

    def test_eq_different_keys(self):
        self.assertFalse(self.Model() == self.Model(id=2))

    def test_eq_same(self):
        self.assertTrue(self.Model(id=1) == self.Model(id=1))

    def test_ne_wrong_type(self):
        self.assertTrue(self.Model() != object())

    def test_ne_wrong_model_type(self):
        class OtherModel(Common, self.Base):
            id = Column(Integer, primary_key=True)
        self.assertTrue(self.Model(id=1) != OtherModel(id=1))

    def test_ne_different(self):
        self.assertTrue(self.Model(id=1) != self.Model(id=2))

    def test_ne_same(self):
        self.assertTrue(self.Model(id=1) != self.Model(id=2))

    def test_repr(self):
        compare('Model(id=1, value=3)', repr(self.Model(id=1, value=3)))

    def test_str(self):
        compare('Model(id=3, value=1)', str(self.Model(id=3, value=1)))

    def test_repr_relationships_excluded(self):
        model = self.AnotherModel(id=1, other=self.Model(id=1, value=42))
        self.session.add(model)
        self.session.flush()
        compare('AnotherModel(id=1, other_id=1)',
                actual=str(model))


class CompareTests(SetupModels, TestCase):

    def check_raises(self, x, y, message, **kw):
        try:
            compare(x, y, **kw)
        except Exception as e:
            if not isinstance(e, AssertionError):
                raise # pragma: no cover
            actual = e.args[0]
            if actual != message: # pragma: no cover
                self.fail(compare(actual, expected=message,
                                  show_whitespace=True,
                                  raises=False))
        else: # pragma: no cover
            self.fail('No exception raised!')

    def test_identical(self):
        compare(
            [self.Model(id=1), self.Model(id=2)],
            [self.Model(id=1), self.Model(id=2)]
            )

    def test_different(self):
        self.check_raises(
            self.Model(id=1), self.Model(id=2),
            "Model not as expected:\n"
            '\n'
            'same:\n'
            "['value']\n"
            "\n"
            "values differ:\n"
            "'id': 1 != 2"
        )

    def test_different_types(self):
        self.check_raises(
            self.Model(id=1), self.AnotherModel(id=1),
            "Model(id=1) != AnotherModel(id=1)"
        )

    def test_db_versus_non_db_equal(self):
        self.session.add(self.Model(id=1))
        self.session.add(self.AnotherModel(id=2, other_id=1))
        self.session.commit()

        db = self.session\
            .query(self.AnotherModel)\
            .options(joinedload(self.AnotherModel.other, innerjoin=True))\
            .one()

        raw = self.AnotherModel(id=2, other_id=1)

        compare(db, raw)

    def test_db_versus_non_db_not_equal(self):
        self.session.add(self.Model(id=1))
        self.session.add(self.AnotherModel(id=2, other_id=1))
        self.session.commit()

        db = self.session\
            .query(self.AnotherModel)\
            .options(joinedload(self.AnotherModel.other, innerjoin=True))\
            .one()

        raw = self.AnotherModel(id=2, other=self.Model(id=2), attr=6)

        self.check_raises(
            db, raw,
            "AnotherModel not as expected:\n"
            "\n"
            "same:\n"
            "['id']\n"
            "\n"
            "values differ:\n"
            "'attr': None != 6\n"
            "'other_id': 1 != None"
        )

    def test_backref_equal(self):
        db = self.AnotherModel(id=2, other=self.Model(id=1, value=2))
        self.session.add(db)
        self.session.commit()
        db = self.session.query(self.AnotherModel).one()
        raw = self.AnotherModel(id=2, other=self.Model(id=1, value=2))
        compare(raw, db, ignore_fields=['other_id'])

    def test_ignore_fields(self):
        compare(self.Model(id=1), self.Model(id=2), ignore_fields=['id'])



from datetime import datetime as dt
from unittest import TestCase

from mortar_rdb.testing import get_session, register_session
from psycopg2.extras import DateTimeRange as Range
from sqlalchemy import CheckConstraint, PrimaryKeyConstraint
from sqlalchemy.dialects.postgresql import ExcludeConstraint
from sqlalchemy.exc import DataError, IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, String
from testfixtures import ShouldRaise, compare

from mortar_mixins.common import Common
from mortar_mixins.temporal import Temporal


class Helper(object):

    @classmethod
    def setUpClass(cls):
        register_session(transactional=False)

    def setUp(self):
        b = declarative_base()
        class Model(Temporal, Common, b):
            key_columns = ['name']
            name = Column(String)
        self.Model = Model
        self.session = get_session()
        self.addCleanup(self.session.rollback)
        b.metadata.create_all(self.session.bind)


class PropertyTests(Helper, TestCase):

    def setUp(self):
        super(PropertyTests, self).setUp()
        self.obj = self.Model(
            name = 'test',
            period = Range(dt(2000, 1, 1), dt(2001, 1, 1))
            )
        
    def test_get_from(self):
        compare(self.obj.value_from, dt(2000, 1, 1))
    
    def test_set_from(self):
        self.obj.value_from = dt(1999, 1, 1)
        compare(self.obj.period,
                Range(dt(1999, 1, 1), dt(2001, 1, 1)))
    
    def test_set_from_none(self):
        self.obj = self.Model()
        self.obj.value_from = dt(1999, 1, 1)
        compare(self.obj.period,
                Range(dt(1999, 1, 1), None))
    
    def test_get_to(self):
        compare(self.obj.value_to, dt(2001, 1, 1))
    
    def test_set_to(self):
        self.obj.value_to = dt(2002, 1, 1)
        compare(self.obj.period,
                Range(dt(2000, 1, 1), dt(2002, 1, 1)))
    
    def test_set_to_none(self):
        self.obj = self.Model()
        self.obj.value_to = dt(1999, 1, 1)
        compare(self.obj.period,
                Range(None, dt(1999, 1, 1)))


class ConstructorTests(Helper, TestCase):

    def test_from(self):
        obj = self.Model(name = 'test', value_from = dt(2000, 1, 1))
        compare(obj.period, Range(dt(2000, 1, 1), None))

    def test_to(self):
        obj = self.Model(name = 'test', value_to = dt(2000, 1, 1))
        compare(obj.period, Range(None, dt(2000, 1, 1)))

    def test_period_and_from(self):
        with ShouldRaise(TypeError(
                'period not allowed if value_from or value_to used'
                )):
            self.Model(
                name = 'test',
                period = Range(dt(2000, 1, 1), dt(2001, 1, 1)),
                value_from = dt(2000, 1, 1),
                )

    def test_period_and_to(self):
        with ShouldRaise(TypeError(
                'period not allowed if value_from or value_to used'
                )):
            self.Model(
                name = 'test',
                period = Range(dt(2000, 1, 1), dt(2001, 1, 1)),
                value_to = dt(2000, 1, 1),
                )


class ValueAtTests(Helper, TestCase):

    def setUp(self):
        super(ValueAtTests, self).setUp()
        # some Models
        self.session.add(self.Model(
                period = Range(dt(2000, 1, 1), dt(2001, 1, 1)),
                name = 'Name 1-1'
                ))
        self.session.add(self.Model(
                period = Range(dt(2001, 1, 1), dt(2002, 1, 1)),
                name = 'Name 1-2'
                ))
        self.session.add(self.Model(
                period = Range(dt(2002, 1, 1), None),
                name = 'Name 1-3'
                ))

    def _check(self, dt, expected):
        objs = self.session.query(self.Model).filter(
            self.Model.value_at(dt)
            ).all()
        actual = [o.name for o in objs]
        compare(expected, actual)
        
    def test_nothing(self):
        self._check(dt(1999, 2, 3), [])
        
    def test_middle(self):
        self._check(dt(2000, 2, 3), ['Name 1-1'])
        self._check(dt(2001, 2, 3), ['Name 1-2'])
        self._check(dt(2002, 2, 3), ['Name 1-3'])

    def test_boundaries(self):
        self._check(dt(2000, 1, 1), ['Name 1-1'])
        self._check(dt(2001, 1, 1), ['Name 1-2'])
        self._check(dt(2002, 1, 1), ['Name 1-3'])


class ConstraintTests(Helper, TestCase):
    
    def _check_valid(self, existing=(), new=(), exception=False):
        # both existing and new should be sequences of two-tuples
        # of the form (value_from,value_to)
        for value_from, value_to in existing:
            self.session.add(self.Model(
                    period = Range(value_from, value_to),
                    name = 'Name'
                    ))

        for value_from, value_to in new:
            self.session.add(self.Model(
                    period = Range(value_from, value_to),
                    name = 'Name'
                    ))
        if exception:
            with ShouldRaise(exception):
                self.session.flush()
        else:
                self.session.flush()

    def test_invalid_1(self):
        # existing:     |---->
        #      new:  |------->
        self._check_valid(
            existing = [(dt(2001, 1, 1), None)],
            new =      [(dt(2002, 1, 1), None)],
            exception = IntegrityError
            )

    def test_invalid_2(self):
        # existing:  |------->
        #      new:     |---->
        self._check_valid(
            existing = [(dt(2001, 1, 1), None)],
            new =      [(dt(2002, 1, 1), None)],
            exception = IntegrityError
            )

    def test_invalid_3(self):
        # existing:  |------->
        #      new:  |------->
        self._check_valid(
            existing = [(dt(2002, 1, 1), None)],
            new =      [(dt(2002, 1, 1), None)],
            exception = IntegrityError
            )

    def test_invalid_4(self):
        # existing:  |-------|
        #      new:  |-------|
        self._check_valid(
            existing = [(dt(2001, 1, 1), dt(2002, 1, 1))],
            new =      [(dt(2001, 1, 1), dt(2002, 1, 1))],
            exception = IntegrityError
            )
        
    def test_invalid_5(self):
        # existing:  |-------|
        #      new:  |-----|
        self._check_valid(
            existing = [(dt(2001, 1, 1), dt(2003, 1, 1))],
            new =      [(dt(2001, 1, 1), dt(2002, 1, 1))],
            exception = IntegrityError
            )

    def test_invalid_6(self):
        # existing:  |-------|
        #      new:    |-----|
        self._check_valid(
            existing = [(dt(2001, 1, 1), dt(2003, 1, 1))],
            new =      [(dt(2002, 1, 1), dt(2003, 1, 1))],
            exception=IntegrityError
            )

    def test_invalid_7(self):
        # existing:  |-------|
        #      new:    |---|
        self._check_valid(
            existing = [(dt(2001, 1, 1), dt(2004, 1, 1))],
            new =      [(dt(2002, 1, 1), dt(2003, 1, 1))],
            exception = IntegrityError
            )

    def test_invalid_8(self):
        # existing:    |---|
        #      new:  |-------|
        self._check_valid(
            existing = [(dt(2002, 1, 1), dt(2003, 1, 1))],
            new =      [(dt(2001, 1, 1), dt(2004, 1, 1))],
            exception=IntegrityError
            )

    def test_invalid_9(self):
        #      new:  |<------|
        self._check_valid(
            new = [(dt(2004, 1, 1), dt(2001, 1, 1))],
            exception=DataError
            )

    def test_invalid_11(self):
        # existing:     |---->
        #      new:  |-------|
        self._check_valid(
            existing = [(dt(2001, 1, 1), None)],
            new =      [(dt(2000, 1, 1), dt(2002, 1, 1))],
            exception = IntegrityError
            )

    def test_invalid_12(self):
        # existing:  |-------|
        #      new:     |---->
        self._check_valid(
            existing = [(dt(2000, 1, 1), dt(2002, 1, 1))],
            new =      [(dt(2001, 1, 1), None)],
            exception = IntegrityError
            )

    def test_invalid_13(self):
        #      new:  |
        self.session.add(self.Model(
                period = Range(dt(2004, 1, 1), dt(2004, 1, 1), '()'),
                name = 'Name'
                ))
        with ShouldRaise(IntegrityError):
            self.session.flush()
        
    def test_invalid_14(self):
        #      new:  null
        self.session.add(self.Model(
                name = 'Name'
                ))
        # SQLAlchemy used to raise an IntegrityError here
        # 1.10 onwards raises a CompileError
        with ShouldRaise():
            self.session.flush()
        
    def test_ok_1(self):
        # existing:  |---|
        #      new:      |--->
        self._check_valid(
            existing = [(dt(2001, 1, 1), dt(2002, 1, 1))],
            new =      [(dt(2002, 1, 1), None)],
            )

    def test_ok_2(self):
        # existing:  |---|
        #      new:      |---|
        self._check_valid(
            existing = [(dt(2001, 1, 1), dt(2002, 1, 1))],
            new =      [(dt(2002, 1, 1), dt(2003, 1, 1))],
            )

    def test_ok_3(self):
        # existing:      |---|
        #      new:  |---|
        self._check_valid(
            existing = [(dt(2002, 1, 1), dt(2003, 1, 1))],
            new =      [(dt(2001, 1, 1), dt(2002, 1, 1))],
            )

    def test_ok_4(self):
        # existing:        |---|
        #      new:  |---|
        self._check_valid(
            existing = [(dt(2002, 1, 1), dt(2003, 1, 1))],
            new =      [(dt(2000, 1, 1), dt(2001, 1, 1))],
            )

    def test_ok_5(self):
        # existing:  |---|
        #      new:        |---|
        self._check_valid(
            existing = [(dt(2000, 1, 1), dt(2001, 1, 1))],
            new =      [(dt(2002, 1, 1), dt(2003, 1, 1))],
            )

    def test_ok_6(self):
        # existing:  |---|   |---|
        #      new:      |---|
        self._check_valid(
            existing = [(dt(2000, 1, 1), dt(2001, 1, 1)),
                        (dt(2002, 1, 1), dt(2003, 1, 1))],
            new =      [(dt(2001, 1, 1), dt(2002, 1, 1))],
            )

    def test_ok_7(self):
        #      new:  <-------|
        self._check_valid(new=[(None, dt(2004, 1, 1))])

    def test_invalid_update_instance(self):
        # add
        obj1 = self.Model(
            period = Range(dt(2000, 1, 1), dt(2001, 1, 1)),
            name = 'Name'
            )
        obj2 = self.Model(
            period = Range(dt(2001, 1, 1), None),
            name = 'Name'
            )
        self.session.add(obj1)
        self.session.add(obj2)
        self.session.flush()
        # change to be invalid
        obj1.period = Range(dt(2000, 1, 1), dt(2002, 1, 1))
        # flush, which should raise exception
        with ShouldRaise(IntegrityError):
            self.session.flush()


class TestExcludeConstraintConstruction(Helper, TestCase):

    def setUp(self):
        Base = declarative_base()
        class Booking(Temporal, Common, Base):
            key_columns = ('hotel', 'room')
            hotel =  Column(String)
            room =  Column(Integer)
        self.Booking = Booking
        self.session = get_session()
        self.addCleanup(self.session.rollback)
        Base.metadata.create_all(self.session.bind)

    def test_valid(self):
        self.session.add(self.Booking(
                hotel='h1', room=1, period=Range(dt(2001, 1, 1),)
                ))
        self.session.add(self.Booking(
                hotel='h1', room=2, period=Range(dt(2001, 1, 1),)
                ))
        self.session.add(self.Booking(
                hotel='h2', room=1, period=Range(dt(2001, 1, 1),)
                ))
        self.session.add(self.Booking(
                hotel='h2', room=2, period=Range(dt(2001, 1, 1),)
                ))
        self.session.flush()

    def test_invalid(self):
        self.session.add(self.Booking(
                hotel='h1', room=1, period=Range(dt(2001, 1, 1), None)
                ))
        self.session.add(self.Booking(
                hotel='h1', room=1, period=Range(None, dt(2001, 1, 2))
                ))


class MethodTests(Helper, TestCase):

    def test_period_str_end(self):
        compare(self.Model(value_to=dt(2001, 1, 1)).period_str(),
                'until 2001-01-01 00:00:00')
    
    def test_period_str_start(self):
        compare(self.Model(value_from=dt(2001, 1, 1)).period_str(),
                '2001-01-01 00:00:00 onwards')
    
    def test_period_str_both(self):
        compare(self.Model(
                period=Range(dt(2000, 1, 1), dt(2001, 1, 1))
                ).period_str(),
                '2000-01-01 00:00:00 to 2001-01-01 00:00:00')

    def test_period_str_both_with_time(self):
        compare(self.Model(
                period=Range(dt(2000, 1, 1, 16), dt(2001, 1, 1, 15))
                ).period_str(),
                '2000-01-01 16:00:00 to 2001-01-01 15:00:00')

    def test_period_str_neither(self):
        compare(self.Model().period_str(),
                'unknown')
    
    def test_period_both_none(self):
        compare(self.Model(period=Range(None, None)).period_str(),
                'always')


class TestNoKeyColumns(Helper, TestCase):

    def setUp(self):
        Base = declarative_base()
        class NoKeys(Temporal, Common, Base):
            col = Column(String)
        self.Model = NoKeys
        self.session = get_session()
        self.addCleanup(self.session.rollback)
        Base.metadata.create_all(self.session.bind)

    def test_valid(self):
        self.session.add(self.Model(col='a', period=Range(dt(2001, 1, 1))))
        self.session.add(self.Model(col='b', period=Range(dt(2001, 1, 1))))
        # you get the helper methods, but no exclude constraint
        self.session.flush()


class TestSingleTableInheritance(Helper, TestCase):

    def setUp(self):
        self.Base = declarative_base()
        class TheTable(Temporal, self.Base):
            __tablename__ = 'stuff'
            __mapper_args__ = dict(
                polymorphic_on='type',
                polymorphic_identity='base',
            )
            key_columns = ['type', 'col']
            type = Column(String)
            col = Column(String)
        self.TheTable = TheTable
        class Model1(TheTable):
            __mapper_args = dict(
                polymorphic_identity='model1',
            )
        class Model2(TheTable):
            __mapper_args = dict(
                polymorphic_identity='model2',
            )
        self.Model1 = Model1
        self.Model2 = Model2
        self.session = get_session()
        self.addCleanup(self.session.rollback)
        self.Base.metadata.create_all(self.session.bind)

    def test_only_two_constraints_created(self):
        # paranoid: only one table
        compare(self.Base.metadata.tables.keys(), expected=['stuff'])
        compare(
            sorted(c.__class__.__name__
                   for c in self.TheTable.__table__.constraints),
            expected=(
                'CheckConstraint', 'ExcludeConstraint', 'PrimaryKeyConstraint'
            )
        )

    def test_add_one_of_each(self):
        self.session.add(self.Model1(col='a', period=Range(dt(2001, 1, 1))))
        self.session.add(self.Model2(col='b', period=Range(dt(2001, 1, 1))))
        # you get the helper methods, but no exclude constraint
        self.session.flush()



# helpers for use in migrations

def migrate_temporal(table):
    # imports here so we don't have a dependency on alembic
    from alembic import op
    import sqlalchemy as sa
    from sqlalchemy.schema import Sequence, CreateSequence

    op.execute(CreateSequence(Sequence(table+"_id_seq")))
    op.add_column(table, sa.Column(
        'id', sa.Integer(), nullable=False,
        server_default=sa.text("nextval('"+table+"_id_seq'::regclass)")
    ))
    op.drop_constraint(table+'_pkey', table_name=table)
    op.create_primary_key(table+'_pkey', table, ['id'])



from psycopg2.extras import DateTimeRange
from sqlalchemy import Column, CheckConstraint
from sqlalchemy import Integer
from sqlalchemy.dialects.postgresql import (
    ExcludeConstraint,
    TSRANGE as Range,
    )
from sqlalchemy.event import listen


class Temporal(object):

    key_columns = None

    def __init__(self, **kw):
        value_from = kw.pop('value_from', None)
        value_to = kw.pop('value_to', None)
        if value_from or value_to:
            if 'period' in kw:
                raise TypeError(
                    'period not allowed if value_from or value_to used'
                    )
            period = DateTimeRange(value_from, value_to)
        else:
            period = kw.pop('period', None)
        super(Temporal, self).__init__(period=period, **kw)

    id = Column(Integer, primary_key=True)
    period = Column(Range(), nullable=False)

    @classmethod
    def value_at(cls, timestamp):
        """
        Returns a clause element that returns the current value
        """
        return cls.period.contains(timestamp)

    @property
    def value_from(self):
        return self.period.lower

    @value_from.setter
    def value_from(self, timestamp):
        if self.period is None:
            upper = None
        else:
            upper = self.period.upper
        self.period = DateTimeRange(timestamp, upper)
        
    @property
    def value_to(self):
        return self.period.upper

    @value_to.setter
    def value_to(self, timestamp):
        if self.period is None:
            lower = None
        else:
            lower = self.period.lower
        self.period = DateTimeRange(lower, timestamp)

    def period_str(self):
        if self.period is None:
            return 'unknown'
        value_from = self.period.lower
        value_to = self.period.upper
        if value_from is None and value_to is None:
            return 'always'
        if self.value_from is None:
            return 'until %s' % self.value_to
        if self.value_to is None:
            return '%s onwards' % self.value_from
        return '%s to %s' % (self.value_from, self.value_to)


def add_constraints(mapper, class_):
    if mapper.polymorphic_map and not mapper.polymorphic_on:
        return
    table = class_.__table__
    if class_.key_columns is not None:
        elements = []
        for col_name in class_.key_columns:
            elements.append((getattr(class_, col_name), '='))
        elements.append(('period', '&&'))
        table.append_constraint(ExcludeConstraint(*elements))
    table.append_constraint(CheckConstraint("period != 'empty'::tsrange"))

listen(Temporal, 'instrument_class', add_constraints, propagate=True)
